#ifndef DEBUG
#include <matrix.h>
#include <mex.h>
#endif

#include <mosek.h>
#include <string.h>
#define EPSILON 1.0e-6
#define MAXK 10
#define MAXDIST 1.0e6

typedef struct {
  double   *X;
  char *isnbr;
  int       D;
  int     num;
} t_nbrStruct;

typedef struct {
  double *Xc;
  int *Isort;
  int      D;
  int      N;
} t_ptsStruct;



typedef struct {
  MSKenv_t         env;
  MSKtask_t       task;
  MSKrescodee        r;
  double          cval;
	MSKlidxt aptrb,aptre;
	MSKidxt         asub;
	double          aval;
	int              bkx;
	double       blx,bux;
	int              bkc;
	double       blc,buc;
  int              Qnz;
} t_msk;

static void MSKAPI printstr(void *handle,
                            char str[])
{
  printf("%s",str);
} /* printstr */


/* Change quicksort to exit early after maxK */
void q_sort(int I[], double numbers[], int left, int right)
{
  int l_hold = left;
  int r_hold = right;
  double pivot = numbers[I[left]];
  int t;
  while (left < right)
  {
    while ((numbers[I[right]] >= pivot) && (left < right))
      right--;
    if (left != right)
    {
      t = I[left]; 
      I[left] = I[right];
      I[right] = t;
      left++;
    }
    while ((numbers[I[left]] <= pivot) && (left < right))
      left++;
    if (left != right)
    {
      t = I[right];
      I[right] = I[left];
      I[left] = t;
      right--;
    }
  }
  numbers[I[left]] = pivot;
  pivot = (double) left;
  left = l_hold;
  right = r_hold;
  if (left < pivot)
    q_sort(I, numbers, left, (int) pivot-1);
  if (right > pivot)
    q_sort(I, numbers, (int) pivot+1, right);
}

void sortK(int I[], double numbers[], int N, int maxK) 
{
  int i,j,mini=0;
  double min;
  for (i=0;i<maxK;i++) {
    min = MAXDIST;
    for (j=0;j<N;j++) {
      if (numbers[j]<min && numbers[j] >= 0.0) {
        mini = j;
        min = numbers[j];
      }
    }
    I[i] = mini;
    numbers[mini] = -1.0;
  }
}

void initmosek(t_msk *msk, int N) 
{  
  /* Make the mosek environment. */
  MSK_makeenv(&(msk->env),NULL,NULL,NULL,NULL);
  MSK_linkfunctoenvstream(msk->env,
      MSK_STREAM_LOG,
      NULL,
      printstr);
  MSK_initenv(msk->env);
  
  msk->cval = 0.0;
	msk->aptrb = 0;
	msk->aptre = 1;
	msk->asub = 0;
	msk->aval = 1.0;
	msk->bkx = MSK_BK_LO;
	msk->blx = 0.0; 
	msk->bux = MSK_INFINITY;
	msk->bkc = MSK_BK_FX;
	msk->blc = 1.0; 
	msk->buc = 1.0;

  /* Make the optimization task. */
  /* msk->r = MSK_maketask(msk->env,1,N,&(msk->task)); */

  /* MSK_putmaxnumqnz(msk->task,(N+1)*N/2); */
}

void appendquad(t_msk *msk, double *Xn, int n, int D) {
  int i,d;
  double qval;
  
#ifdef DEBUG
  double *dbg_qval = malloc(sizeof(double)*n);
  int numvar;
#endif

  MSK_appendvars (msk->task,1,
	      &msk->cval,
        &msk->aptrb,&msk->aptre,
        &msk->asub,&msk->aval,
        &msk->bkx,&msk->blx,&msk->bux); 

  for (i=0;i<=n;i++) {
    qval = 0;
    for (d=0;d<D;d++) {
       qval = Xn[(n-1)*D+d]*Xn[i*D+d] + qval;              
    } 
#ifdef DEBUG
  dbg_qval[i] = qval;
#endif
    if (qval > EPSILON || qval < -EPSILON) {
      msk->Qnz++;     
      MSK_putqobjij (msk->task,n-1,i,qval);
    }
  }  
#ifdef DEBUG
  printf("Qnz=%i\n",msk->Qnz);
  printf("Q=");
  for (i=0;i<n;i++) {
    printf("%1.4f, ",dbg_qval[i]);
  }
  printf("\n");
  MSK_getnumvar(msk->task, &numvar);
  /* free(dbg_qval); */
#endif

}


double
solvequad(t_msk *msk, double *X, double *xx, double *xproj, int D, int N)
{
  int i,d;
  double y = 0.0;

  MSK_optimize(msk->task);

  MSK_getsolutionslice(msk->task,
    MSK_SOL_ITR,
    MSK_SOL_ITEM_XX,
    0,
    N,
    xx);

  for(d=0;d<D;d++) {
    xproj[d] = 0;
    for (i=0;i<N;i++) {    
      xproj[d] = xproj[d] + xx[i]*X[i*D+d];
    }
    y = y + xproj[d]*xproj[d];
  }

#ifdef DEBUG
  printf("xproj=");
  for (d=0;d<D;d++) {
    printf("%1.4f, ",xproj[d]);
  }
  printf("\n");

  printf("y: %1.4f\n",y);
  printf("N: %i\n",N);

  printf("xx=");
  for (i=0;i<N;i++) {
    printf("%1.4f, ",xx[i]);
  }
  printf("\n");
#endif

  return y;
}

void initquad(t_msk *msk, const double *X0, int D, int N) {
  int   d;
  double qval = 0;

  msk->r = MSK_maketask(msk->env,1,N,&(msk->task));

  MSK_putmaxnumqnz(msk->task,(N+1)*N/2);

  msk->r = MSK_inputdata(msk->task,1,1,1,1,
          &msk->cval,0.0,
          &msk->aptrb,&msk->aptre,
          &msk->asub,&msk->aval,
          &msk->bkc,&msk->blc,&msk->buc,
          &msk->bkx,&msk->blx,&msk->bux);

  for (d=0;d<D;d++) {
    qval = X0[d]*X0[d] + qval;              
  } 

  if (qval < EPSILON && qval > -EPSILON) {
    msk->Qnz = 0;
  }
  else {
    msk->Qnz = 1;  
    MSK_putqobjij(msk->task,0,0,qval);
  }

#ifdef DEBUG
  printf("Q=%1.4f\n",qval);
#endif
}

void getCenterSort(const double *x, const double *X, double *Xcs, int *I, int maxK, int D, int N) 
{
  int d,i;
  double *dist = malloc(sizeof(double)*N);
  double Xc;

  /* Center the X's and compute the L2 norm */
  for (i=0;i<N;i++) {
    dist[i] = 0.0;
    for (d=0;d<D;d++) {
      Xc = X[i*D+d] - x[d];
      dist[i] = dist[i] + Xc*Xc;  
    }
  }

  sortK(I, dist, N, maxK); 

  free(dist);

  for (i=0;i<maxK;i++) {
    for (d=0;d<D;d++) {
      Xcs[i*D+d] = X[I[i]*D+d] - x[d];
    }
  }

  #ifdef DEBUG
    printf("I=");
    for (i=0;i<maxK;i++) {
      printf("%i, ",I[i]+1);
    }
    printf("\n");

    printf("dist=");
    for (i=0;i<maxK;i++) {
      printf("%1.4f, ",dist[I[i]]);
    }
    printf("\n");

    printf("Xcs=");
    for (i=0;i<D*maxK;i++) {
      printf("%1.4f, ",Xcs[i]);
    }
    printf("\n");
  #endif
}

double addNeighbor(t_nbrStruct *nbrs, const double *Xc, int pos, int D) 
{
  int d;
  double dist = 0.0;

  for (d=0;d<D;d++) {
    nbrs->X[nbrs->num*D+d] = Xc[pos*D+d];
    dist = dist + Xc[pos*D+d]*Xc[pos*D+d];
  }
  nbrs->num++;
  nbrs->isnbr[pos] = 1;

  return dist;
}

int getNextToCheck(const double *xproj, const double* Xc, double dist, const char *isnbr, int D, int N)
{
  int i,d;
  double proj;

#ifdef DEBUG
  printf("xproj=");
  for (d=0;d<D;d++) {
    printf("%1.4f, ",xproj[d]);
  }
  printf("\n");
#endif

  for (i=1;i<N;i++) {
    if (!isnbr[i]) {
      proj = 0;
      for (d=0;d<D;d++) {
        proj = proj + xproj[d]*Xc[i*D+d];  
      }     
      if (proj < dist + EPSILON) { 
#ifdef DEBUG
  printf("Next Neighbor=%i\n",i);
#endif
        return i;
      }
    }
  }
  return -1;
}

int getSENk_centered(t_msk *msk, const double *Xc, int minK, int D, int N) 
{
  int i;
  double *xx = malloc(sizeof(double)*N);
  /* int *I = malloc(sizeof(int)*N); */
  double *xproj = malloc(sizeof(double)*D);
  int nextToCheck;  
  double dist;
  int maxDistNeighborIndex = 0;

  /* Make a struct to hold neighbor info */
  double *Xnbrs = malloc(sizeof(double)*N*D); 
  char *isnbr = malloc(sizeof(char)*N); /* (int)(N/8+1)); */
  t_nbrStruct nbrs; 
  nbrs.X = Xnbrs;
  nbrs.isnbr = isnbr;
  nbrs.num = 0;

    /*Reset the neighbors */
    for (i=0;i<N;i++) {
      nbrs.isnbr[i] = 0;
    }
    nbrs.num = 0;

  /*Set the nearest point as a neighbor */
  dist = addNeighbor(&nbrs, Xc, 0, D);

#ifdef DEBUG
  printf("Nbrs=");
  for (i=0;i<nbrs.num*D;i++) {
    printf("%1.4f,",nbrs.X[i]);
  }
  printf("\n");
  printf("\nDist=%1.1f \n",dist);
#endif

  initquad(msk,nbrs.X,D,N);

  /*Check the next point on the near side of the hyperplane */
  nextToCheck = getNextToCheck(&Xc[0], Xc, dist, nbrs.isnbr, D, N);
  
  /*Include only the remaining points on the near side of hyperplane */
  while (nextToCheck != -1) {  
    if (nextToCheck > maxDistNeighborIndex) {
      maxDistNeighborIndex = nextToCheck;
    }

    /*Add as a neighbor */
    addNeighbor(&nbrs, Xc, nextToCheck, D);

    /*Find the distance to the convex hull */
    appendquad(msk,nbrs.X,nbrs.num,D);
    dist = solvequad(msk, nbrs.X, xx, xproj, D, nbrs.num);


#ifdef DEBUG
    /*printf("maxDistNeighborIndex=%i\n",maxDistNeighborIndex); */
    printf("Nbrs=");
    for (i=0;i<nbrs.num*D;i++) {
      printf("%1.4f,",nbrs.X[i]);
    }
    printf("\n");

    printf("\nDist=%e \n",dist);
#endif

    /*If you are in the hull OR all points are neighbors, quit. */
    if (dist < EPSILON || nbrs.num == N) {
      break;
    }    

    nextToCheck = getNextToCheck(xproj, Xc, dist, nbrs.isnbr, D, N);
  }

  MSK_deletetask(&(msk->task));

  free(xx);
  free(xproj);

  free(Xnbrs);
  free(isnbr);

  return maxDistNeighborIndex; /*These are your k-1 nearest neighbors */
}

int getSENk(double *k, int **Isort, const double *x, const double *X, double *maxK, double *minK, int D, int N, int M) 
{
  int m;
  int Kmax = (int) *maxK; 
  int Kmin = (int) *minK;
  double *Xc = malloc(sizeof(double)*D*Kmax);
  int totK = 0;
  int *currPos;
  t_msk msk;

  /*int *Isort = malloc(sizeof(int)*N*Kmax); */
  *Isort = malloc(sizeof(int)*M*Kmax);
  
  initmosek(&msk,Kmax);

  for (m=0;m<M;m++) {
    currPos = *Isort + totK;

    getCenterSort(&x[m*D], X, Xc, currPos, Kmax, D, N);

    k[m] = (double) getSENk_centered(&msk,Xc,Kmin,D,Kmax) + 1;

    if (k[m] < Kmin) { 
      k[m] = Kmin;
    }

    totK = totK + (int) k[m];

    #ifdef DEBUG
      printf("k=%i\n\n\n\n",(int)k[m]);
    #endif
  }      

  MSK_deleteenv(&(msk.env));
  free(Xc);
  /*free(Isort); */
  return totK;
}


#ifndef DEBUG
void
mexFunction(int nlhs, mxArray *plhs[],
            int nrhs, const mxArray *prhs[])
{

  const mxArray *X, *x, *minK, *maxK;
  mxArray *k;
  mxArray *Imatlab;
  int D,N,M;
  int totK;
  int *I;

  X = prhs[0];
  x = prhs[1];
  minK = prhs[2];
  maxK = prhs[3];

  N = mxGetN(X);
  D = mxGetM(X);
  M = mxGetN(x);

  /*printf("\n%i\n",M); */

  if (mxGetM(x) != D)
    mexErrMsgTxt("X and x must have the same number of rows!");

  k = mxCreateNumericMatrix(M,1,mxDOUBLE_CLASS,mxREAL);

  /*printf("ok!\n"); */

  totK = getSENk(mxGetPr(k), &I, mxGetPr(x), mxGetPr(X), mxGetPr(maxK), mxGetPr(minK), D, N, M);

  Imatlab = mxCreateNumericMatrix(totK,1,mxINT32_CLASS,mxREAL);
  memcpy ((int*) mxGetData(Imatlab), I, sizeof(int)*totK);
  free(I);

  plhs[0] = k;
  plhs[1] = Imatlab;
  
}
#else

int main(void)
{
  /*
  double X[] = {1, 1.1, 2, 0, 0, 1, 3, 4, 5, 5, -10, -10, -20, 15};
  double x[] = {0,0};
  int D = 2;
  int N = 7;

  double X[] = {-1.2,0.2,-3.4,-0.8,0.4,-0.2,0.0,0.8,0.3,-2.1,0.4,0.5,-0.6,-0.0,1.5,0.2,0.0,2.0,-0.1,-1.3,0.2,-0.4,-0.5,-0.2,-1.5,-0.8,-0.7,0.1,0.4,-0.8};
  double x[] = {0,0,0};
  int D = 2;
  int N = 15;   

  double X[] = {-1.4592,1.5797,-0.6107,-1.9593,1.5515,0.0660,0.4602,0.0895,-0.0394,0.4964,-0.8649,1.0598,-0.5365,-1.1517,0.1606,-0.1679,1.1615,-1.3345,-0.6557,0.5635,};
  double x[] = {0.5316,0.4797};
  int D = 2;
  int N = 10;  
 */
  
  double X[] = {-0.7015,-0.1666,0.1163,-0.7145,0.7573,-0.0765,-0.6739,0.1495,-1.0242,-1.6968,0.8340,-1.6435,0.1504,0.8699,0.5801,-1.0721,-0.7938,-0.0168,0.0932,-0.2174};
  double x[] = {-0.4935, 0.8354, 0, 0};
  int D = 2;
  int N = 10;  
  
  int M = 2;
  int i,j;
  double* k = malloc(sizeof(double)*M);
  double mink = 4;
  double maxk = (double) N;
  int totK;
  int *I;
  int run;

  /*(double *k, const double *x, const double *X, double *maxK, double *minK, int D, int N, int M) */ 

  totK = getSENk(k, &I, &x[0], X, &maxk, &mink, D, N, M);

  printf("totK=%i\n",totK);
  
  run = 0;
  for(i=0;i<M;i++) {
    printf("I[%i]=",i);
    for(j=0;j<k[i];j++) {
      printf("%i,",I[run+j]);
    }
    printf("\n");
    run = run+(int)k[i];
  }
  
  /*printf("k=%i\n",(int)k); */
  /*scanf("\n"); */

  free(I);
  free(k);
}
#endif
