% REGRESS_LOCALLINEAR performs local linear regression
%   Y2 = REGRESS_LOCALLINEAR(X1,Y1,X2,NHOOD,KLIM,REGL,LAMBDA)
%   Regresses the functional trend in the training data (X1,Y1) to predict
%   the outputs for test inputs X2.
%     X1 -- NxD matrix of training inputs (N samples, D dimensions)
%     Y1 -- NxP matrix of training outputs (N samples, P dimensions)
%     X2 -- MxD matrix of test inputs (M samples, D dimensions)
%     Y2 -- MxP matrix of test outputs (M samples, P dimensions)
%
%   Additional parameters control the type of local linear regression:
%     NEIGHBORHOOD STYLE_____________________________________________
%     NHOOD = 'knn' - Normal KNN  | 'eknn' - Enclosing KNN
%     KLIM  = [k]                 | [smallest_k, largest_k]
%
%     REGULARIZATION_________________________________________________
%     REGL  = {'none','ridge','tikhonov'}
%     LAMBDA = Regularization Parameter (may be empty if REGL='none').
% 
%   Example
%       X1=rand(100,3); Y1=prod(X1,2); X2=rand(20,3);
%       regress_LocalLinear(X1,Y1,X2,'eknn',[3,30],'tikhonov',1);
function Y2 = regress_LocalLinear(X1,Y1,X2,nhood,Klim,regl,lambda)
  time.start = clock;  

  [nY1 dY] = size(Y1);    
  [nX2 dX] = size(X2);

  if strcmp(regl,'tikhonov'),
    beta0 = myLinear(Y1,X1);
    beta0 = beta0(2:end,:);
  end
  
  %Loop through all the X2 points
  Y2 = zeros(nX2,dY);    
  for i = 1:nX2
    %Set the neighbors
    if strcmp(nhood,'prespecified')
      if size(Klim,1) ~= nX2, 
        error('prespecified knn dimension mismatch!'); 
      end
      I = Klim(i,:);
      I = I(I~=0);
    else
      I = findNeighbors(X1, X2(i,:), nhood, Klim);
    end    
    fixed = std(X1(I,:)) < 1e-3;   
    Y = Y1(I,:);
    X = X1(I,~fixed);
    
    %Use specified regression function
    switch lower(regl)
      case 'none'
        beta = myLinear(Y,X);
      case 'ridge'
        beta = myRidge(Y,X,lambda);
      case 'tikhonov'    
        mybeta0 = beta0(~fixed,:);
        beta = myTikhonov(Y,X,lambda,mybeta0);
      otherwise
        error('unknown regularization');
    end      
    
    %Solve for the output
    Y2(i,:) = [1 X2(i,~fixed)]*beta;    
  end
  
%   Y2 = Y2';
  time.elap = etime(clock,time.start);
  %fprintf('local linear regression (%s) took %1.3fs\n',regl,time.elap);
end

function beta = myLinear(Y,X)
  [nX,dX] = size(X);
  [nY,dY] = size(Y);
  
  mx   = mean(X);
  stdx = std(X,0,1);  
  X = (X-repmat(mx,[nX,1]) ) ./ repmat(stdx,[nX,1]);    
  
  beta = zeros(dX+1,dY);  
  for j=1:dY
    y = Y(:,j);
    b = (X'*X)\(X'*y)./stdx'; 
    beta(:,j) = [mean(y) - mx*b; b];
  end
end

function beta = myRidge(Y,X,lambda)
  [nX,dX] = size(X);
  [nY,dY] = size(Y);
 
  mx   = mean(X);
  stdx = std(X,0,1);  
  X = (X-repmat(mx,[nX,1]) ) ./ repmat(stdx,[nX,1]);  
  
  beta = zeros(dX+1,dY);  
  for j=1:dY
    y = Y(:,j);
    b = (X'*X + lambda*eye(dX))\(X'*y)./stdx'; 
    beta(:,j) = [mean(y) - mx*b; b];
  end    
end

function beta = myTikhonov(Y,X,lambda,b0)  
  [nX,dX] = size(X);
  [nY,dY] = size(Y);
  
  mx   = mean(X);
  stdx = std(X,0,1);  
  
  X = (X-repmat(mx,[nX,1]) ) ./ repmat(stdx,[nX,1]);
  b0 = repmat(stdx',[1,dY]).*b0;
 
  beta = zeros(dX+1,dY);  
  for j=1:dY
    y = Y(:,j);
    b = (X'*X + lambda*eye(dX))\(X'*y + lambda*b0(:,j))./stdx'; 
    beta(:,j) = [mean(y) - mx*b; b];
  end    
end

function I = findNeighbors(X,x,nhood,Klim)
  nX = size(X,1);
  
  %Set the regression function
  switch lower(nhood)
    case 'knn'
      if strcmpi(Klim,'all')
        nNbrs = nX;
      else  
        nNbrs = min(Klim(1),nX);
      end      
      Dx = sum( ( X - repmat(x,[nX,1]) ).^2 , 2);
      [dum,I] = sort(Dx,'ascend');  
      I = I(1:nNbrs); 
    case 'eknn'
      eknn = getEnclosingKNN_qp(X,x,Klim);
      I = eknn.indices(eknn.indices>0);
    otherwise
      error('unknown neighborhood option!');
  end    
end
