%GETGRID Get a mesh of n^d points spanning lim(i,1) to lim(i,2) for
%dimension i.
%   GETGRID(N,LIM) given size(LIM) = [D,2], and size(N) = [D,1] returns 
%   a D x (N(1)*N(2)*...*N(D)) matrix of gridded samples of the rectangle 
%   spanning lim(i,1) to lim(i,2) in dimension i. There are N(i) points 
%   along dimension i. Entries of N must be a non-negative integers smaller 
%   than 2^52 and size(LIM,1) must be an integer between 2 and 36.
%
%   Examples
%       getGrid([2,2],[0,1;0,1])  %returns [0 0 1 1
%                     %                 0 1 0 1];
%       getMesh([3,2],[0,1;0,1])  %returns [0.0 0.0 0.0 0.5 0.5 0.5 1.0 1.0 1.0
%                                       0.0 0.5 1.0 0.0 0.5 1.0 0.0 0.5 1.0];
%
%   Derived from DEC2BASE (The MathWorks, Inc).
function varargout = getGrid(varargin)

switch nargin
  case 1
    %Sending in a grid, get its dimensions
    X2 = varargin{1};
    [dG,nG] = size(X2);
    
    nGrid = zeros(1,dG);
    skip = 1;
    for i=1:dG-1
      D = diff(X2(i,1:skip:end),1);
      [dum,I] = max(abs(D(2:end)));
      nGrid(i) = I+1;
      skip = nGrid(i)*skip;
    end
    nGrid(dG) = length(X2(dG,1:skip:end));
    
    varargout{1} = nGrid;
    
  case 2
    %Requesting a grid, create it.
    n = varargin{1};
    lim = varargin{2};
    
    %repeat n if it's a scalar
    if length(n) == 1
      n = repmat(n,[1,size(lim,1)]);
    end

    if any(size(lim)~=1)
      d = size(lim,1);
    else
      d = lim;
      lim = repmat([0 1],[d,1]);
    end

    %mesh the domain and scale from the unit mesh
    x = getMesh(n);  
    scale = @(x) x.*repmat(diff(lim,[],2),[1,size(x,2)]) + ...
                   +repmat(lim(:,1),[1,size(x,2)]); 
    g = scale(x);
    
    varargout{1} = g;    
end



  
function s = getMesh(n)
dim = length(n);
if nargin<1
    error(nargchk(1,2,nargin,'struct'));
end
if ~(isnumeric(n) || ischar(n)) || any(n ~= floor(n)) || any(n < 0) || any(n > 1/eps)
    error('getMesh:FirstArg', 'First argument must be an integer, 0 <= n <= 2^52.');
end
% if ~isscalar(dim) || ~(isnumeric(dim) || ischar(dim)) || dim ~= floor(dim) || dim < 1 || dim > 36
%     error('getMesh:SecondArg', 'Second argument must be an integer, 1 <= dim <= 36.');
% end

d = double(0:prod(n)-1);
% b = double(n);

s = zeros(prod(n),dim);
for i=1:dim
  b = double(n(i));
  s(:,i) = rem(d,b)';
  d = floor(d/b); 
end
s = (s./repmat(n-1,[prod(n),1]))';