function S = getLinearInterpolationMatrix(x,g)

  %collect some information
  [dX, nX] = size(x);
  [dG, nG] = size(g); 
  if dX~=dG, error('x is dimension %d while grid is dimension %d',dX,dG); end

  %center and scale the inputs to 0-1 for convienience
  limGrid = [min(g,[],2), max(g,[],2)];
  scale   = @(x) ( x - repmat(limGrid(:,1),[1,size(x,2)]) )...
                 ./ repmat(diff(limGrid,[],2),[1,size(x,2)]); 
  g = scale(g);
  x = scale(x);   
  
  %Make sure the input is within range of the grid
  if any(x(:)>1) || any(x(:)<0)
    limGrid    
    error('\nThe input points are not within your LUT grid');
  end 
  
  %Extract the size of the grid
  nGrid = getGrid(g); 
  
  %Compute the principal cell index and the lsb in each dimension
  nGrid_x   = repmat(nGrid',[1,nX]);
  binI      = min( floor(x.*(nGrid_x-1)), nGrid_x-2 );
  lsb       = x.*(nGrid_x-1) - binI;
  
  %Compute the indices of all corners of the cell containing each x.
  bin_ramp  = getBinaryRamp(dX);
  index_mod = [1 cumprod(nGrid(1:end-1))];
  offset    = bin_ramp * index_mod';
  bin       = index_mod * binI + 1;
  I         = repmat(bin,[2^dX,1]) + repmat(offset,[1,nX]);  

  %Compute the interpolation weights for each x
  bin_ramp = repmat(permute(bin_ramp,[1,3,2]),[1,nX,1]);
  lsb      = repmat(permute(lsb,[3 2 1]),[2^dX,1,1]);
  vals     = prod(bin_ramp.*lsb + ~bin_ramp.*(1-lsb),3);          

  %Now build the matrix that performs the linear interpolation
  cols = repmat(1:nX, [2^dX,1]);
  S = sparse(I(:), cols(:), vals(:), prod(nGrid), nX);      
  
%   %Debug Plotting-------------------
%   figure(1); clf; 
%   axis equal; set(gca,'nextplot','add');
%   if dX==2
%     for i=1:size(g,2)
%       plot(g(1,i),g(2,i),'.');
%       text(g(1,i),g(2,i),num2str(i));
%     end
%     for i=1:size(x,2)
%       plot(x(1,i),x(2,i),'g.');
%       text(x(1,i),x(2,i),num2str(i));
%     end
%   elseif dX==3
%     for i=1:size(g,2)
%       plot3(g(1,i),g(2,i),g(3,i),'.');
%       text(g(1,i),g(2,i),g(3,i),num2str(i));
%     end
%     for i=1:size(x,2)
%       plot3(x(1,i),x(2,i),x(3,i),'g.');
%       text(x(1,i),x(2,i),x(3,i),num2str(i));
%     end    
%   end
%   %---------------------------------  
end

function s = getBinaryRamp(dim)
  n = repmat(2,[1,dim]);
  d = double(0:prod(n)-1);
  s = zeros(prod(n),dim);
  for i=1:dim
    b = double(n(i));
    s(:,i) = rem(d,b)';
    d = floor(d/b); 
  end
end

