function hview=plotLUT(grid,varargin) 
%PLOTFUNCTION plots a lut function from GRID
%   Examples for plotFunction:
% 
%   %2D data
%   f = gmm(2*(rand(2,10)-.25),.1);
%   grid = newLUT(17,2);
%   grid.y = f(grid.x);
%     %(Without) reference points
%     plotLUT(grid); 
%     %(With)    reference points
%     train.x = rand(2,10);
%     train.y = f(train.x);
%     plotLUT(grid,train);
% 
%   %3D data
%   f = gmm(2*(rand(3,10)-.25),.1);
%   grid = newLUT(17,3);
%   grid.y = f(grid.x);
%   lim = [min(grid.y(:)) max(grid.y(:))];
%   values = [0:.05:1]*(lim(2)-lim(1))+lim(1);
%     %(Without) reference points
%     figure(1), plotLUT(grid,values);
%     %(With)    reference points
%     train.x = rand(3,10);
%     train.y = f(train.x);
%     plotLUT(grid,values,train);

  [d,nX] = size(grid.x);
  shape = grid.n;

  clf; view(3);  
  set(gca,'nextplot','add');  
  
  %Two dimensional plotting (easy!)
  if d==2
    hview = hggroup;    
    
    if nargin == 2,
      train = varargin{1};
      hs    = scatter3(train.x(1,:),train.x(2,:),train.y,'filled');
      set(hs,'parent',hview);      
    end
    
    X = reshape(grid.x(1,:),shape);
    Y = reshape(grid.x(2,:),shape); 
    Z = reshape(grid.y,shape);    
    surf(X,Y,Z,'linestyle','none','facealpha',.5,'Parent',hview);
       
  %Three dimensional plotting (harder!)
  elseif d==3      
    if nargin < 2, error('Enter values to be plotted for 3d data'); end
    values = varargin{1}; 
      
    %extract some useful information about the values to be plotted
    step = values(2)-values(1);
    lim  = [values(1) - step, values(end) + step];

    %create a function that colors points based on their value
    C = jet;
    range = (0:1/(size(C,1)-1):1)*(lim(2)-lim(1)) + lim(1);
    clip = @(x) x - (x>255).*(x-255) - (x<0).*(x);
    r = @(val,dim) clip(interp1(range,C(:,dim),val,'linear','extrap'));
    color = @(val) [r(val,1) r(val,2) r(val,3)];      
    
    %Plot all the training points
    if nargin == 3 
      train  = varargin{2};
      scatter3(train.x(2,:),train.x(3,:),train.x(1,:),10,...
        color(train.y'),'filled');
    else
      train = [];
    end

    %Format the grid for an isosurface
    X = reshape(grid.x(2,:),shape); xlabel('dim 2');
    Y = reshape(grid.x(3,:),shape); ylabel('dim 3');
    Z = reshape(grid.x(1,:),shape); zlabel('dim 1');
%     X = reshape(grid.x(1,:),shape);  
%     Y = reshape(grid.x(2,:),shape);
%     Z = reshape(grid.x(3,:),shape);
    V = reshape(grid.y,shape);    
    
    hview = {};  
    for i=1:length(values);
      val = values(i);
     
      %If there are no points in this range, then skip.
      if ~any(V(:)<val) || ~any(V(:)>val)
        continue;
      end
      
      %create a group to hold this step in the animation
      hg = hggroup('UserData',sprintf('V=%1.3f',val));    
      
      %patch the volume      
      hp = patch(isosurface(X,Y,Z,V,val),'Parent',hg,'FaceColor',...
            color(val),'EdgeColor','none','EdgeAlpha',0,'FaceAlpha',.25);
      
      %scatter the nearby (in value) points large-size
      if ~isempty(train)    
        I = abs(val - train.y) <= step/2;
        x = train.x(:,I); y = train.y(:,I);      
        hs = scatter3(x(2,:),x(3,:),x(1,:),50,color(y'),'filled');
        set(hs,'parent',hg);
      end

      hview{end+1} = hg;
    end
  
    minmax = @(A) [min(A(:)) max(A(:))];
    set(gca,'xlim',minmax(X),'ylim',minmax(Y),'zlim',minmax(Z));
    HVScroller(gca,hview,{});
  end
  
end